@@ -3,9 +3,10 @@ This module contains the logic to execute tasks and their dependencies
 to completion.
 """
 
-from concurrent.futures import wait, FIRST_COMPLETED
 import logging
-import time
+from concurrent.futures import wait, FIRST_COMPLETED, ALL_COMPLETED
+
+import attr
 
 from .task_execution import (
     EntryLevel,
@@ -54,11 +55,9 @@ class TaskCompletionRunner:
     enum, indicating whether we need to and/or can do more work on it.
     """
 
-    def __init__(self, bootstrap, flow_instance_uuid, task_key_logger):
+    def __init__(self, context):
         # These are needed to complete entries.
-        self._bootstrap = bootstrap
-        self._flow_instance_uuid = flow_instance_uuid
-        self.task_key_logger = task_key_logger
+        self._context = context
 
         # These are used for caching and tracking.
         self._entries_by_task_key = {}
@@ -68,25 +67,23 @@ class TaskCompletionRunner:
 
     @property
     def _parallel_execution_enabled(self):
-        if self._bootstrap is None:
-            return False
-        return self._bootstrap.process_executor is not None
+        return self._context.core.process_executor is not None
 
     @property
     def _aip_execution_enabled(self):
-        if self._bootstrap is None:
-            return False
-        return self._bootstrap.aip_executor is not None
+        return self._context.core.aip_executor is not None
 
     def run(self, states):
         try:
             if self._parallel_execution_enabled:
-                self._bootstrap.process_executor.start_logging()
+                self._context.core.process_executor.start_logging()
 
             for state in states:
                 entry = self._get_or_create_entry_for_state(state)
                 self._add_requirement(
-                    src_entry=None, dst_entry=entry, level=EntryLevel.CACHED
+                    dst_entry=entry,
+                    level=EntryLevel.CACHED,
+                    expiration=EntryRequirement.Expiration.NEVER,
                 )
 
             while self._has_pending_entries():
@@ -102,13 +99,13 @@ class TaskCompletionRunner:
             for state in states:
                 task_key = state.task_key
                 entry = self._entries_by_task_key[task_key]
-                result = entry.get_cached_result(self.task_key_logger)
+                result = entry.get_cached_result(self._context)
                 results.append(result)
             return results
 
         finally:
             if self._parallel_execution_enabled:
-                self._bootstrap.process_executor.stop_logging()
+                self._context.core.process_executor.stop_logging()
 
     def _process_active_entry(self, entry):
         """
@@ -133,12 +130,17 @@ class TaskCompletionRunner:
         # so we'll need them to be primed.
         self._set_up_entry_dependencies(entry)
         for dep_entry in entry.dep_entries:
-            self._add_requirement(entry, dep_entry, EntryLevel.PRIMED)
+            self._add_requirement(
+                src_entry=entry,
+                dst_entry=dep_entry,
+                level=EntryLevel.PRIMED,
+                expiration=EntryRequirement.Expiration.WHEN_SRC_CACHED,
+            )
         if self._mark_entry_blocked_if_necessary(entry):
             return
 
         # Now that we have the dependency digests, we can initialize our task state.
-        entry.state.initialize(self._bootstrap, self._flow_instance_uuid)
+        entry.state.initialize(self._context)
 
         # If that was all we needed, we're done.
         if self._mark_entry_completed_if_possible(entry):
@@ -166,7 +168,7 @@ class TaskCompletionRunner:
             self._aip_execution_enabled or self._parallel_execution_enabled
         ) and entry.state.should_persist
         if entry_may_be_computable_remotely:
-            remote_subgraph = RemoteSubgraph(entry.state, self._bootstrap)
+            remote_subgraph = RemoteSubgraph(entry.state)
             if self._aip_execution_enabled:
                 aip_task_configs = remote_subgraph.distinct_aip_task_configs
             else:
@@ -178,7 +180,7 @@ class TaskCompletionRunner:
                 descriptions_str = "; ".join(
                     f"function outputting {state.task_key.dnode.to_descriptor()!r} "
                     f"requires {state.func_attrs.aip_task_config}"
-                    for state in remote_subgraph.stripped_states_with_aip_task_configs
+                    for state in remote_subgraph.states_with_aip_task_configs
                 )
                 message = f"""
                 Multiple functions need to be run together (since some are not
@@ -194,11 +196,11 @@ class TaskCompletionRunner:
                     # as an AIP-decorated derived node.
                     non_serializable_descriptors = [
                         state.task_key.dnode.to_descriptor()
-                        for state in remote_subgraph.non_serializable_stripped_states
+                        for state in remote_subgraph.non_serializable_states
                     ]
                     aip_descriptors = [
                         state.task_key.dnode.to_descriptor()
-                        for state in remote_subgraph.stripped_states_with_aip_task_configs
+                        for state in remote_subgraph.states_with_aip_task_configs
                     ]
                     message = f"""
                     Found impossible configuration:
@@ -222,23 +224,45 @@ class TaskCompletionRunner:
             entry_is_computable_remotely = False
 
         if entry_is_computable_remotely:
-            # When we run an entry remotely, we also need to run all of its immediate
-            # non-persistable ancestors, since their values can't be shared between
-            # processes. Those ancestors may have follow-up tasks, so we'll need to
-            # run those too; and if those follow-ups are persistable (but not actually
-            # persisted yet), their values will be usable everywhere, so we'll want to
-            # track the fact that they're running. So really we're going to run a
-            # collection of persistable entries (including the original entry), plus
-            # their immediate non-persistable ancestors.
+            # To run an entry remotely, we need to transmit and compute a subgraph of
+            # tasks. This subgraph will contain one or more "target task states" that
+            # will be computed together, along with "dependency task states" that need
+            # to already be computed and persisted. See the documentation of the
+            # RemoteSubgraph class for more info on how this subgraph is constructed.
             target_entries = [
-                self._get_or_create_entry_for_state(persistable_subgraph_state)
-                for persistable_subgraph_state in remote_subgraph.persistable_but_not_persisted_states
+                self._get_or_create_entry_for_state(target_state)
+                for target_state in remote_subgraph.target_states
             ]
             assert entry in target_entries
 
-            # We want all of our entries to be initialized before we try running
-            # them remotely. We'll just require this of the persistable ones, since
-            # all the other ones are ancestors of them.
+            dep_entries = [
+                self._get_or_create_entry_for_state(external_dep_state)
+                for external_dep_state in remote_subgraph.external_dependency_states
+            ]
+            assert entry not in dep_entries
+
+            # All of our dependency entries will need to be cached before we can run
+            # this subgraph. (One might think that they should already be cached because
+            # they are persistable and they are ancestors of the active entry, so we
+            # should have cached all of them by now. However, it's possible that some
+            # of their artifacts have been deleted since then, so we need to make sure
+            # to actually instantiate their entries and confirm that they're still
+            # cached.)
+            for dep_entry in dep_entries:
+                assert dep_entry.state.should_persist
+                self._add_requirement(
+                    src_entry=entry,
+                    dst_entry=dep_entry,
+                    level=EntryLevel.CACHED,
+                    expiration=EntryRequirement.Expiration.WHEN_SRC_CACHED,
+                )
+            # We want our target states and their ancestors to be initialized before we
+            # run this remotely.
+            # TODO Is this requirement still necessary?
+            # (Note that this doesn't guarantee that every state in the subgraph is
+            # initialized; there can still be non-persistable followups that don't get
+            # initialized here. This isn't a problem for those states, since their
+            # remote execution doesn't affect anything in this process.)
             for target_entry in target_entries:
                 if target_entry == entry:
                     assert target_entry.level >= EntryLevel.INITIALIZED
@@ -247,25 +271,26 @@ class TaskCompletionRunner:
                         src_entry=entry,
                         dst_entry=target_entry,
                         level=EntryLevel.INITIALIZED,
+                        expiration=EntryRequirement.Expiration.WHEN_SRC_CACHED,
                     )
             if self._mark_entry_blocked_if_necessary(entry):
                 return
 
+            active_target_entries = []
             for target_entry in target_entries:
-                # Now we'll also require that each entry be CACHED, which guarantees
-                # that it won't be COMPLETED. (If we didn't do this, the entry's
-                # required level might only be INITIALIZED, in which case it would
-                # already be COMPLETED. However, we know it's not already CACHED,
-                # because it's `persistable_but_not_persisted`.)
+                # Now, we potentially have multiple target entries to compute at once.
+                # We'll add a CACHED requirement to indicate that we want to compute
+                # them, and then we'll activate the ones that aren't cached yet.
                 self._add_requirement(
-                    src_entry=None,
                     dst_entry=target_entry,
                     level=EntryLevel.CACHED,
+                    expiration=EntryRequirement.Expiration.WHEN_MET,
                 )
                 if target_entry.stage == EntryStage.PENDING:
                     self._mark_entry_active(target_entry)
-                # The entry should now be ACTIVE, because:
-                # - It can't be COMPLETED because of the requirement we added above.
+
+                # At this point the entry should be either COMPLETED (if it's already
+                # cached) or ACTIVE (if it's not).
                 # - It can't be PENDING because we would have just activated it above.
                 # - It shouldn't be BLOCKED, because:
                 #   - we already required the entry to be at least INITIALIZED, and
@@ -276,55 +301,76 @@ class TaskCompletionRunner:
                 #    our original entry, then the reverse should also be true (it's a
                 #    symmetric relationship), so our original entry would already be
                 #    IN_PROGRESS or COMPLETED too.
-                assert target_entry.stage == EntryStage.ACTIVE
+                if target_entry.stage == EntryStage.ACTIVE:
+                    active_target_entries.append(target_entry)
+                else:
+                    assert target_entry.stage == EntryStage.COMPLETED
 
-            stripped_target_states = [
-                remote_subgraph.get_stripped_state(target_entry.state)
-                for target_entry in target_entries
-            ]
-            new_bootstrap = self._bootstrap.evolve(
-                aip_executor=None, process_executor=None
+            stripped_active_target_states = remote_subgraph.strip_states(
+                active_target_entry.state
+                for active_target_entry in active_target_entries
             )
-            new_task_completion_runner = TaskCompletionRunner(
-                bootstrap=new_bootstrap,
-                flow_instance_uuid=self._flow_instance_uuid,
-                task_key_logger=self.task_key_logger,
+
+            # Remove the executors when sending the jobs over so that the remote
+            # executor does not attempt to create its own subprocess or AIP
+            # jobs.
+            new_core = self._context.core.evolve(
+                aip_executor=None,
+                process_executor=None,
+            )
+            new_context = self._context.evolve(
+                core=new_core,
+                temp_result_cache=MemoryResultCache(),
             )
             if aip_task_config is not None:
-                future = self._bootstrap.aip_executor.submit(
+                new_context = new_context.evolve(
+                    task_key_logger=TaskKeyLogger(new_core),
+                )
+                new_task_completion_runner = TaskCompletionRunner(new_context)
+                future = self._context.core.aip_executor.submit(
+                    entry.state.task_key,
                     aip_task_config,
                     run_in_subprocess,
                     new_task_completion_runner,
-                    stripped_target_states,
+                    stripped_active_target_states,
                 )
 
                 def done_callback(callback_future):
-                    if not callback_future.cancelled():
-                        for target_entry in target_entries:
-                            self.task_key_logger.log_computed_aip(
+                    if (
+                        not callback_future.cancelled()
+                        and callback_future.exception() is None
+                    ):
+                        for target_entry in active_target_entries:
+                            self._context.task_key_logger.log_computed_aip(
                                 target_entry.state.task_key
                             )
 
                 future.add_done_callback(done_callback)
             else:
-                future = self._bootstrap.process_executor.submit(
+                new_task_completion_runner = TaskCompletionRunner(new_context)
+                future = self._context.core.process_executor.submit(
                     run_in_subprocess,
                     new_task_completion_runner,
-                    stripped_target_states,
+                    stripped_active_target_states,
                 )
 
-            for target_entry in target_entries:
+            for target_entry in active_target_entries:
                 self._mark_entry_in_progress(target_entry, future)
             return
 
         # Otherwise we'll compute this entry locally. In that case, it's not enough
         # for our dependencies to be primed; they also need to have cached values.
         for dep_entry in entry.dep_entries:
-            self._add_requirement(entry, dep_entry, EntryLevel.CACHED)
+            self._add_requirement(
+                src_entry=entry,
+                dst_entry=dep_entry,
+                level=EntryLevel.CACHED,
+                expiration=EntryRequirement.Expiration.WHEN_SRC_CACHED,
+            )
         if self._mark_entry_blocked_if_necessary(entry):
             return
 
-        entry.compute(self.task_key_logger)
+        entry.compute(self._context)
         assert self._mark_entry_completed_if_possible(entry)
 
     def _set_up_entry_dependencies(self, entry):
@@ -336,8 +382,22 @@ class TaskCompletionRunner:
             for dep_state in entry.state.dep_states
         ]
 
-    def _add_requirement(self, src_entry, dst_entry, level):
-        req = EntryRequirement(src_entry=src_entry, dst_entry=dst_entry, level=level)
+    def _add_requirement(
+        self,
+        dst_entry,
+        level,
+        expiration,
+        src_entry=None,
+    ):
+        if expiration == EntryRequirement.Expiration.WHEN_SRC_CACHED:
+            assert src_entry is not None
+
+        req = EntryRequirement(
+            dst_entry=dst_entry,
+            level=level,
+            expiration=expiration,
+            src_entry=src_entry,
+        )
         if req.src_entry is not None:
             req.src_entry.outgoing_reqs.add(req)
         req.dst_entry.incoming_reqs.add(req)
@@ -370,8 +430,11 @@ class TaskCompletionRunner:
             return self._entries_by_task_key[task_key]
         # Before doing anything with this task state, we should make sure its
         # cache state is up to date.
-        state.refresh_all_persistent_cache_state(self._bootstrap)
-        entry = TaskRunnerEntry(state)
+        # TODO Having to clear the state with each get() call is brittle. It would
+        # probably be better to store this cache/artifact state in the ExecutionContext
+        # so it will automatically get thrown away at the end of the query.
+        state.refresh_all_persistent_cache_state(self._context)
+        entry = TaskRunnerEntry(self._context, state)
         self._entries_by_task_key[task_key] = entry
         return entry
 
@@ -403,32 +466,62 @@ class TaskCompletionRunner:
 
     def _wait_on_in_progress_entries(self):
         "Waits on any in-progress entry to finish."
-        futures = set(entry.future for entry in self._in_progress_entries)
-        if self._parallel_execution_enabled:
-            finished_futures, _ = wait(futures, return_when=FIRST_COMPLETED)
-        else:
-            assert self._aip_execution_enabled
-            # TODO: This works for PoC, but maybe there is a better way
-            # to do this. Also consider making the sleep time configurable.
-            # See this comment for example
-            # https://github.com/square/bionic/pull/253#discussion_r487066911
-            while all(future.running() for future in futures):
-                time.sleep(10)
-            finished_futures = [future for future in futures if future.done()]
+        entries_by_future = {entry.future: entry for entry in self._in_progress_entries}
+        finished_futures, _ = wait(
+            entries_by_future.keys(), return_when=FIRST_COMPLETED
+        )
         for finished_future in finished_futures:
-            # TODO: If there is an error, consider waiting till all
-            # futures are done. With AIP execution, we can have one
-            # task fail while others are running. If we don't wait till
-            # those tasks are done, the user might perform another
-            # `get` operation that can spawn the same Task again. Apart
-            # from wasted resource, the other problem is that the older
-            # and newer tasks can potentially conflict writing to the
-            # same cache file.
-            task_keys = finished_future.result()
-            for task_key in task_keys:
-                entry = self._entries_by_task_key[task_key]
-                entry.state.sync_after_remote_computation()
-                assert self._mark_entry_completed_if_possible(entry)
+            try:
+                self._sync_and_complete_remotely_computed_task_keys(
+                    finished_future.result()
+                )
+            except Exception as exception:
+                # If there is an error, wait until all futures are done. With
+                # AIP execution, we can have one task fail while others are
+                # running. If we don't wait till those tasks are done, the user
+                # might perform another `get` operation that can spawn the same
+                # Task again.
+
+                logger.exception(
+                    oneline(
+                        f"""
+                        Encountered an error while doing remote computation for
+                        {entries_by_future[finished_future].state.task_key}.
+                        Waiting for all other remote computation(s) to
+                        complete...
+                        """
+                    )
+                )
+                finished_futures, _ = wait(
+                    entries_by_future.keys(), return_when=ALL_COMPLETED
+                )
+
+                # There may be errors in the other tasks. Ensure that we log all
+                # of the other errors as well.
+                for future in finished_futures:
+                    try:
+                        self._sync_and_complete_remotely_computed_task_keys(
+                            future.result()
+                        )
+                    except Exception as e:
+                        if e != exception and e is not None:
+                            logger.exception(
+                                oneline(
+                                    f"""
+                                    Encountered another error while doing remote
+                                    computation for
+                                    {entries_by_future[future].state.task_key}.
+                                    """
+                                )
+                            )
+
+                raise exception
+
+    def _sync_and_complete_remotely_computed_task_keys(self, task_keys):
+        for task_key in task_keys:
+            entry = self._entries_by_task_key[task_key]
+            entry.state.sync_after_remote_computation()
+            assert self._mark_entry_completed_if_possible(entry)
 
     def _mark_entry_completed_if_possible(self, entry):
         assert entry.stage in [EntryStage.ACTIVE, EntryStage.IN_PROGRESS]
@@ -448,41 +541,121 @@ class TaskCompletionRunner:
             entry.future = None
             self._in_progress_entries.remove(entry)
 
+        entry.stage = EntryStage.COMPLETED
+
         if entry.level >= EntryLevel.CACHED:
             # If we have any followup tasks, we want to run them immediately.
             for followup_state in entry.state.followup_states:
                 followup_entry = self._get_or_create_entry_for_state(followup_state)
-                self._raise_entry_priority(followup_entry, EntryPriority.HIGH)
+                # If this entry's value is a collection of values, then we want to
+                # run its followups as soon as possible in order to get the
+                # (potentially large) object out of memory.
+                if entry.state.desc_metadata.is_composite:
+                    priority = EntryPriority.TOP
+                # Otherwise, we still want to make sure followups run before any
+                # non-followups.
+                else:
+                    priority = EntryPriority.HIGH
+                self._raise_entry_priority(followup_entry, priority)
                 self._add_requirement(
-                    src_entry=None,
                     dst_entry=followup_entry,
                     level=EntryLevel.CACHED,
+                    expiration=EntryRequirement.Expiration.WHEN_MET,
+                )
+                # We also add a requirement from the followup back to its parent.
+                # This requirement will get generated anyway as soon as we start
+                # processing the followup, but we want to add it early so the
+                # requirement is visible when we start pruning requirements below;
+                # otherwise the current entry might think it has no more requirements
+                # and evict its value from memory.
+                #
+                # (Normally every entry we process was requested by one of its children,
+                # so when it completes it always has at least one active requirement.
+                # However, followups are requested by their parents and get processed
+                # before their children, so it's possible for a followup to complete
+                # with no active. That's why we need to add this requirement early.)
+                assert entry.state in followup_entry.state.dep_states
+                self._add_requirement(
+                    src_entry=followup_entry,
+                    dst_entry=entry,
+                    level=EntryLevel.CACHED,
+                    expiration=EntryRequirement.Expiration.WHEN_SRC_CACHED,
                 )
 
+            # Now this this entry has reached the final level of progress, some
+            # requirements may be ready to expire.
+            self._prune_requirements_for_cached_entry(entry)
+
+            # And now that some requirements have been removed, we may be able to
+            # evict some memoized values from memory.
             for dep_entry in entry.dep_entries:
-                self._vacate_completed_entry_if_possible(dep_entry)
-
-        entry.stage = EntryStage.COMPLETED
-
-        # TODO This might be a good place to prune old, already-met requirements.
+                # For each temporarily-memoized input value, we can try to evict it
+                # *unless* it has other high-priority (i.e., followup) entries waiting
+                # for it. (In general, we want these values to be recomputed each time
+                # they're needed, *except* for composite values like tuples, which need
+                # to stick around long enough to be broken up by their followup tassks.)
+                self._evict_temp_memoized_entry_value_if_possible(
+                    dep_entry, ignore_normal_priority_reqs=True
+                )
+            # We can also try to evict our own value if it has no active requirements
+            # at all. (This will generally happen if this entry was created as a
+            # followup but has no followups of its own. Otherwise, if we do have any
+            # requirements, we want the value to stick around long enough to actually
+            # get used by one of them!)
+            self._evict_temp_memoized_entry_value_if_possible(entry)
 
         return True
 
-    def _vacate_completed_entry_if_possible(self, entry):
-        assert entry.stage == EntryStage.COMPLETED
+    def _prune_requirements_for_cached_entry(self, entry):
+        assert entry.level >= EntryLevel.CACHED
 
-        # We only vacate an entry if it has followup entries.
-        if len(entry.state.followup_states) == 0:
+        # This entry is cached, so all incoming requirements must be already be met. If
+        # that matches their expiration criterion, we can discard them.
+        incoming_reqs_to_discard = [
+            req
+            for req in entry.incoming_reqs
+            if req.expiration == EntryRequirement.Expiration.WHEN_MET
+        ]
+        # Similary, all outgoing requirements have this entry as a source, so we know
+        # their source is cached.
+        outgoing_reqs_to_discard = [
+            req
+            for req in entry.outgoing_reqs
+            if req.expiration == EntryRequirement.Expiration.WHEN_SRC_CACHED
+        ]
+
+        for req in incoming_reqs_to_discard + outgoing_reqs_to_discard:
+            req.dst_entry.incoming_reqs.discard(req)
+            if req.src_entry is not None:
+                req.src_entry.outgoing_reqs.discard(req)
+
+    def _evict_temp_memoized_entry_value_if_possible(
+        self, entry, ignore_normal_priority_reqs=False
+    ):
+        if not self._context.temp_result_cache.contains(entry.state.task_key):
             return
 
-        # We only vacate an entry once all its followups are cached; at that point,
-        # we know we won't need this one's value again.
-        for followup_state in entry.state.followup_states:
-            followup_entry = self._get_or_create_entry_for_state(followup_state)
-            if followup_entry.level != EntryLevel.CACHED:
-                return
+        # If this is enabled, the value should be saved until this query completes.
+        if entry.state.should_memoize_for_query:
+            return
 
-        entry.vacate()
+        remaining_reqs = [
+            req for req in entry.incoming_reqs if req.level >= EntryLevel.CACHED
+        ]
+        if ignore_normal_priority_reqs:
+            remaining_reqs = [
+                req
+                for req in remaining_reqs
+                if req.src_entry is not None
+                and req.src_entry.priority > EntryPriority.NORMAL
+            ]
+        if len(remaining_reqs) > 0:
+            return
+
+        self._context.temp_result_cache.delete(entry.state.task_key)
+
+        if entry.stage == EntryStage.COMPLETED and not entry.all_incoming_reqs_are_met:
+            self._mark_entry_pending(entry)
 
     def _mark_entry_blocked_if_necessary(self, entry):
         assert entry.stage == EntryStage.ACTIVE
@@ -526,25 +699,16 @@ class TaskKeyLogger:
     of a computation.)
     """
 
-    def __init__(self, bootstrap):
-        self._level = logging.INFO if bootstrap is not None else logging.DEBUG
+    def __init__(self, core):
+        self._level = core.task_key_logging_level
 
-        executor = bootstrap.process_executor if bootstrap is not None else None
-        if executor is not None:
+        if core.process_executor is not None:
             self._already_logged_entity_case_key_pairs = (
-                executor.create_synchronized_set()
+                core.process_executor.create_synchronized_set()
             )
         else:
-            # TODO: When AIP execution is enabled, we will send this
-            # SynchronizedSet to AIP jobs for logging task keys. AIP
-            # will get the already added keys to the set, but any
-            # updates made to the set in AIP won't be returned back to
-            # the main process. This will results in progress reporting
-            # being inaccurate, since any job computed in AIP would not
-            # be logged and will be logged as loaded from disk in the
-            # main process as a result. We should fix AIP progress
-            # reporting by either logging the keys in the main process
-            # or rethinking how we do progress reporting.
+            # In AIP execution, any logs and updates to this set are not
+            # returned back to the main bionic process.
             self._already_logged_entity_case_key_pairs = SynchronizedSet()
 
     def _log(self, template, task_key, is_resolved=True):
@@ -592,7 +756,49 @@ class TaskKeyLogger:
         self._log("Computed   %s", task_key)
 
     def log_computed_aip(self, task_key):
-        self._log("Computed   %s using AIP", task_key)
+        self._log("Computed   %s using AI Platform", task_key)
+
+
+# TODO Consider introducing the term "query" to refer to the scope of one get() call.
+@attr.s(frozen=True)
+class ExecutionContext:
+    """
+    Holds objects common to a specific "execution" -- i.e., a single ``flow.get()``
+    call. This is a convenience class to save the trouble of passing these objects
+    around individually.
+    """
+
+    flow_instance_uuid = attr.ib()
+    core = attr.ib()
+    task_key_logger = attr.ib()
+    # This is used for temporarily saving results for the duration of one execution.
+    # Currently we only do this for entities where both persistence and memoization are
+    # disabled.
+    temp_result_cache = attr.ib()
+
+    def evolve(self, **kwargs):
+        return attr.evolve(self, **kwargs)
+
+
+class MemoryResultCache:
+    """
+    A simple cache for storing computed Results by their TaskKeys.
+    """
+
+    def __init__(self):
+        self._results_by_task_key = {}
+
+    def contains(self, task_key):
+        return task_key in self._results_by_task_key
+
+    def save(self, result):
+        self._results_by_task_key[result.task_key] = result
+
+    def load(self, task_key):
+        return self._results_by_task_key.get(task_key)
+
+    def delete(self, task_key):
+        del self._results_by_task_key[task_key]
 
 
 def dnode_without_drafts(dnode):
