@@ -18,11 +18,12 @@ in the correct order; thus, the TaskRunnerEntry mostly contains data pertaining
 task's relationship to other tasks.
 """
 
-import attr
 import copy
-import logging
-
 from enum import auto, Enum, IntEnum
+import logging
+from typing import Optional
+
+import attr
 
 from ..datatypes import Artifact, Result
 from ..exception import (
@@ -32,7 +33,7 @@ from ..exception import (
 )
 from ..persistence import Provenance, ProvenanceDigest
 from ..utils.files import ensure_parent_dir_exists
-from ..utils.misc import oneline
+from ..utils.misc import hash_simple_obj_to_hex, oneline
 from ..utils.urls import path_from_url, url_from_path
 
 logger = logging.getLogger(__name__)
@@ -60,11 +61,10 @@ class TaskRunnerEntry:
         TaskCompletionRunner.
     """
 
-    def __init__(self, state):
+    def __init__(self, context, state):
+        self.context = context
         self.state = state
         self.future = None
-        self.result = None
-        self.is_vacated = False
         self._stage = None
         self.stage = EntryStage.COMPLETED
 
@@ -74,7 +74,7 @@ class TaskRunnerEntry:
 
         self.incoming_reqs = set()
         self.outgoing_reqs = set()
-        self.priority = EntryPriority.DEFAULT
+        self.priority = EntryPriority.NORMAL
 
     @property
     def stage(self):
@@ -116,12 +116,18 @@ class TaskRunnerEntry:
 
     @property
     def _is_cached(self):
-        return self.result is not None or self.state.is_cached
+        if self.state.should_persist:
+            return self.state.is_cached
+        else:
+            return (
+                self.context.temp_result_cache.contains(self.state.task_key)
+                or self.state.is_cached
+            )
 
     @property
     def _is_primed(self):
         if self.state.should_persist:
-            return self.state.is_cached
+            return self._is_cached
         else:
             return self.state.is_initialized
 
@@ -129,7 +135,7 @@ class TaskRunnerEntry:
     def _is_initialized(self):
         return self.state.is_initialized
 
-    def compute(self, task_key_logger):
+    def compute(self, context):
         """
         Computes the value of an entry by running its task. Requires that all
         the task's dependencies are already computed.
@@ -140,20 +146,21 @@ class TaskRunnerEntry:
 
         state = self.state
         task = state.task
-        provenance = state._provenance
         protocol = state.desc_metadata.protocol
 
         assert state.is_initialized
         assert not state.is_cached
 
+        assert task is not None, (state.task_key, self.level)
+
         dep_results = []
         for dep_entry, dep_key in zip(self.dep_entries, task.dep_keys):
             assert dep_entry._is_cached
-            dep_result = dep_entry.get_cached_result(task_key_logger)
+            dep_result = dep_entry.get_cached_result(context)
             dep_results.append(dep_result)
 
         if not task.is_simple_lookup:
-            task_key_logger.log_computing(state.task_key)
+            context.task_key_logger.log_computing(state.task_key)
 
         dep_values = [dep_result.value for dep_result in dep_results]
 
@@ -161,7 +168,6 @@ class TaskRunnerEntry:
         if state.output_would_be_missing():
             result = Result(
                 task_key=task.key,
-                provenance=provenance,
                 value=None,
                 local_artifact=None,
                 value_is_missing=True,
@@ -183,86 +189,43 @@ class TaskRunnerEntry:
         value = task.compute(dep_values)
 
         if task.is_simple_lookup:
-            task_key_logger.log_accessed_from_definition(state.task_key)
+            context.task_key_logger.log_accessed_from_definition(state.task_key)
         else:
-            task_key_logger.log_computed(state.task_key)
+            context.task_key_logger.log_computed(state.task_key)
 
         protocol.validate_for_dnode(task.key.dnode, value)
         result = Result(
             task_key=task.key,
-            provenance=provenance,
             value=value,
             local_artifact=None,
         )
 
         if state.should_persist:
-            artifact = state._local_artifact_from_value(result.value)
+            artifact = state._local_artifact_from_value(result.value, context)
             state._cache_accessor.save_local_artifact(artifact)
             state._result_value_hash = artifact.content_hash
 
         # If we're not persisting the result, this is our only chance to memoize it;
-        # otherwise, we can memoize it later if/when we load it get_cached_result.
+        # otherwise, we can memoize it later if/when we load it from get_cached_result.
         # (It's important to memoize the value we loaded, not the one we just computed,
         # because they may be subtly different and we want all downstream tasks to get
         # exactly the same value.)
         elif state.should_memoize:
             state._result = result
-        else:
-            self.result = result
 
-    def get_cached_result(self, task_key_logger):
+        else:
+            self.context.temp_result_cache.save(result)
+
+    def get_cached_result(self, context):
         "Returns the result of an already-computed entry."
 
         assert self._is_cached
 
-        if self.is_vacated:
-            descriptor = self.state.task_key.dnode.to_descriptor()
-            message = f"""
-            Attempted to access a memoized value for {descriptor} after it was vacated;
-            this should never happen unless you're using an undocumented descriptor
-            feature; otherwise, this is probably a bug in Bionic.
-            """
-            raise AssertionError(oneline(message))
+        result = self.context.temp_result_cache.load(self.state.task_key)
+        if result is not None:
+            return result
 
-        if self.result is not None:
-            return self.result
-        return self.state.get_cached_result(task_key_logger)
-
-    def vacate(self):
-        """
-        Deletes any result memoized on this entry.
-
-        The purpose is to avoid holding values in memory if we know they won't be
-        needed. This is used for tuple-producing entries once all their followups
-        have been completed. We know we won't need the tuple values anymore (see NOTE
-        below), and we want the downstream entries to be the ones to decide if the
-        values stay in memory or not.
-
-        Does not affect anything memoized on the TaskState -- only on this entry itself.
-
-        NOTE We assume tuple entries will never be requested directly by user code,
-        so the only entries that need to access their values are the
-        automatically-created followup tasks. This assumption is not actually safe if
-        the user uses tuple descriptors, but these aren't currently documented. Once
-        we have file descriptors, the only tasks with followups will be the "crude"
-        pre-persisted values, which we won't document and which the user will have no
-        reason to access. Furthermore, at that point we should be able to simplify
-        the task architecture, so it may be easier to implement this "vacation" idea
-        more gracefully, such as by allowing an entry to become un-CACHED and then
-        get recomputed again.
-        """
-
-        if self.result is None:
-            return
-
-        self.is_vacated = True
-        # Instead of setting this to None, we set it to a dummy string. If we just set
-        # it to None, then _is_cached would return False and it would look like this
-        # entry had become un-CACHED, which could cause confusing behavior. This way,
-        # the entry will continue be CACHED, but is_vacated will cause us to throw an
-        # exception if anyone tries to actually access the cached values (which should
-        # never happen).
-        self.result = "SHOULD NEVER BE ACCESSED"
+        return self.state.get_cached_result(context)
 
     def __str__(self):
         return f"TaskRunnerEntry({self.state.task_key})"
@@ -320,16 +283,33 @@ class EntryRequirement:
     """
     Represents a requirement from one entry to another.
 
-    A requirement indicates that one entry (`src_entry`) can't make any further progress
-    until another entry (`dst_entry`) has reached a certain level of progress (`level`).
+    A requirement indicates that we need a particular entry (``dst_entry``) to reach a
+    certain level of progress (``level``). Once reached, this level needs to be
+    maintained until a certain point (``expiration``); afterwards the requirement is no
+    longer in effect and can be ignored (and deleted).
 
-    Note: `src_entry` can also be `None`, indicating some sort of external or a priori
-    requirement. On the other hand, `dst_entry` cannot be `None`.
+    There are three possible values for ``expiration``:
+
+    - ``WHEN_MET``: When ``dst_entry` reaches ``level``.
+    - ``WHEN_SRC_CACHED``: When another entry (``src_entry``) reaches the ``CACHED``
+        stage. Generally ``src_entry`` will depend on the requirement being met, so this
+        condition implies the previous one.
+    - ``NEVER``: This requirement is in effect forever.
     """
 
-    src_entry = attr.ib()
-    dst_entry = attr.ib()
-    level = attr.ib()
+    class Expiration(IntEnum):
+        """
+        Represents the point when a requirement ceases to be in effect.
+        """
+
+        WHEN_MET = auto()
+        WHEN_SRC_CACHED = auto()
+        NEVER = auto()
+
+    dst_entry: TaskRunnerEntry = attr.ib()
+    level: "EntryLevel" = attr.ib()
+    expiration: Expiration = attr.ib()
+    src_entry: Optional[TaskRunnerEntry] = attr.ib()
 
     @property
     def is_met(self):
@@ -375,6 +355,10 @@ class EntryLevel(IntEnum):
         in the persistent cache, in memory on the TaskState, and/or in memory on this
         entry (depending on the cache settings). This is the final level: after this,
         there is no more work to do on this task.
+
+    Normally an entry will only make forward progress through these levels; however,
+    we do sometimes evict temporarily-memoized values, which can cause an entry to
+    regress from CACHED to PRIMED.
     """
 
     CREATED = auto()
@@ -388,16 +372,23 @@ class EntryPriority(IntEnum):
     Indicates a level of priority for a TaskRunnerEntry.
 
     When multiple entries are in the PENDING stage, an entry with higher priority will
-    always be activated before one with lower priority. There are currently only two
-    priorities: DEFAULT and HIGH.
+    always be activated before one with lower priority. There are currently three
+    priorities:
 
-    Currently, the only reason to prioritize one entry over another is because we may
-    want one entry's output to be saved to a persistent cache before another entry has
-    a chance to run (and possibly crash).
+    1. NORMAL: Most entries will have this priority.
+    2. HIGH: This is for entries that some meaningful side effect depends on. For
+       example, after computing a value, we want to make sure it gets persisted to disk
+       (if appropriate) before any other work happens.
+    3. TOP: This is for entries that some effect depends on *and* that depends on a
+       potentially large in-memory objet. For example, if we compute a tuple value, our
+       first priority should be to decompose it into smaller objects, allowing the
+       original tuple to be garbage-collected; until that happens, the individual
+       objects can't be garbage-collected either.
     """
 
-    DEFAULT = auto()
+    NORMAL = auto()
     HIGH = auto()
+    TOP = auto()
 
 
 # TODO Let's reorder the methods here with this order:
@@ -452,6 +443,11 @@ class TaskState:
         self._result_value_hash = None
 
         # This can be set by get_cached_result() or compute().
+        # TODO It would be nice to move this to a central in-memory cache object, like
+        # context.temp_result_cache but with a longer lifetime. However, it would be
+        # a little weird to move this but still have self._result_value_hash here.
+        # Would it makes sense to remove the latter altogether and just retrieve it
+        # lazily from self._cache_accessor?
         self._result = None
 
     @property
@@ -459,26 +455,24 @@ class TaskState:
         return self.desc_metadata.should_memoize
 
     @property
-    def should_persist(self):
-        return self.desc_metadata.should_persist and not self.output_would_be_missing()
+    def should_memoize_for_query(self):
+        return self.desc_metadata.should_memoize_for_query
 
     @property
-    def should_cache(self):
-        return self.should_memoize or self.should_persist
+    def should_persist(self):
+        return self.desc_metadata.should_persist and not self.output_would_be_missing()
 
     @property
     def is_cached(self):
         """
         Indicates whether the task state's result is cached.
         """
-        return self._result_value_hash is not None or self._result is not None
-
-    @property
-    def is_cached_persistently(self):
-        """
-        Indicates whether the task state's result is cached persistently.
-        """
-        return self._result_value_hash is not None
+        if self.should_persist:
+            # If our value is persistable, it can be saved either on disk or in memory,
+            # but only the former counts as being officially "cached".
+            return self._result_value_hash is not None
+        else:
+            return self._result is not None
 
     def output_would_be_missing(self):
         return self.task_key.case_key.has_missing_values
@@ -486,25 +480,24 @@ class TaskState:
     def __repr__(self):
         return f"TaskState({self.task!r})"
 
-    def get_cached_result(self, task_key_logger):
+    def get_cached_result(self, context):
         "Returns the result of an already-computed task state."
 
         assert self.is_cached
 
         if self._result is not None:
-            task_key_logger.log_accessed_from_memory(self.task_key)
+            context.task_key_logger.log_accessed_from_memory(self.task_key)
             return self._result
 
         local_artifact = self._cache_accessor.replicate_and_load_local_artifact()
         value = self._value_from_local_artifact(local_artifact)
         result = Result(
             task_key=self.task_key,
-            provenance=self._provenance,
             value=value,
             local_artifact=local_artifact,
         )
 
-        task_key_logger.log_loaded_from_disk(result.task_key)
+        context.task_key_logger.log_loaded_from_disk(result.task_key)
 
         if self.should_memoize:
             self._result = result
@@ -528,7 +521,7 @@ class TaskState:
 
         self._load_value_hash()
 
-    def refresh_all_persistent_cache_state(self, bootstrap):
+    def refresh_all_persistent_cache_state(self, context):
         """
         Refreshes all state that depends on the persistent cache.
 
@@ -541,7 +534,7 @@ class TaskState:
         if not self.is_initialized or not self.should_persist:
             return
 
-        self.refresh_cache_accessor(bootstrap)
+        self.refresh_cache_accessor(context)
 
         # If we haven't loaded anything from the cache, we can stop here.
         if self._result_value_hash is None:
@@ -581,22 +574,13 @@ class TaskState:
         if self._result_value_hash is None:
             self._load_value_hash()
 
-    def initialize(self, bootstrap, flow_instance_uuid):
+    def initialize(self, context):
         "Initializes the task state to get it ready for completion."
 
         if self.is_initialized:
             return
 
         # First,  set up the provenance.
-        if bootstrap is None:
-            # If we're still in the bootstrap resolution phase, we don't have
-            # any versioning policy, so we don't attempt anything fancy.
-            treat_bytecode_as_functional = False
-        else:
-            treat_bytecode_as_functional = (
-                bootstrap.versioning_policy.treat_bytecode_as_functional
-            )
-
         dep_provenance_digests_by_task_key = {
             dep_key: dep_state._get_digest()
             for dep_key, dep_state in zip(self.task.dep_keys, self.dep_states)
@@ -606,18 +590,20 @@ class TaskState:
             task_key=self.task_key,
             code_fingerprint=self.func_attrs.code_fingerprint,
             dep_provenance_digests_by_task_key=dep_provenance_digests_by_task_key,
-            treat_bytecode_as_functional=treat_bytecode_as_functional,
+            treat_bytecode_as_functional=(
+                context.core.versioning_policy.treat_bytecode_as_functional
+            ),
             can_functionally_change_per_run=self.func_attrs.changes_per_run,
-            flow_instance_uuid=flow_instance_uuid,
+            flow_instance_uuid=context.flow_instance_uuid,
         )
 
         # Lastly, set up cache accessors.
         if self.should_persist:
-            self.refresh_cache_accessor(bootstrap)
+            self.refresh_cache_accessor(context)
 
         self.is_initialized = True
 
-    def refresh_cache_accessor(self, bootstrap):
+    def refresh_cache_accessor(self, context):
         """
         Initializes the cache acessor for this task state.
 
@@ -627,11 +613,11 @@ class TaskState:
         in order to wipe this state and allow it get back in sync with the real world.
         """
 
-        self._cache_accessor = bootstrap.persistent_cache.get_accessor(
+        self._cache_accessor = context.core.persistent_cache.get_accessor(
             task_key=self.task_key,
             provenance=self._provenance,
         )
-        if bootstrap.versioning_policy.check_for_bytecode_errors:
+        if context.core.versioning_policy.check_for_bytecode_errors:
             self._check_accessor_for_version_problems()
 
     def _check_accessor_for_version_problems(self):
@@ -743,7 +729,7 @@ class TaskState:
             assert self._provenance is not None
             return ProvenanceDigest.from_provenance(self._provenance)
 
-    def _local_artifact_from_value(self, value):
+    def _local_artifact_from_value(self, value, context):
         protocol = self.desc_metadata.protocol
 
         dir_path = self._cache_accessor.generate_unique_local_dir_path()
@@ -768,13 +754,40 @@ class TaskState:
                 )
             ) from e
 
-        value_hash = protocol.tokenize_file(value_path)
+        value_hash = self._generate_value_hash(value, value_path, context)
         return Artifact(
             url=url_from_path(value_path),
             content_hash=value_hash,
         )
         return value_path
 
+    def _generate_value_hash(self, value, value_path, context):
+        protocol = self.desc_metadata.protocol
+        value_hash = protocol.tokenize_file(value_path)
+
+        def _extra_value_hash():
+            code_versioning_policy = self.func_attrs.code_versioning_policy
+            code_version = code_versioning_policy.version
+            versioning_policy = context.core.versioning_policy
+            if (
+                not code_version.includes_bytecode
+                or not versioning_policy.treat_bytecode_as_functional
+            ):
+                return ""
+            protocol = self.desc_metadata.protocol
+            try:
+                return protocol.get_extra_value_hash(
+                    value,
+                    code_versioning_policy.suppress_bytecode_warnings
+                    or versioning_policy.ignore_bytecode_exceptions,
+                )
+            except Exception:
+                if versioning_policy.ignore_bytecode_exceptions:
+                    return ""
+                raise
+
+        return hash_simple_obj_to_hex([value_hash, _extra_value_hash()])
+
     def _value_from_local_artifact(self, local_artifact):
         file_path = path_from_url(local_artifact.url)
         protocol = self.desc_metadata.protocol
@@ -799,84 +812,121 @@ class RemoteSubgraph:
     process).
 
     Given a target TaskState, this class identifies the minimal set of TaskStates that
-    should be run along with it; this includes all of its immediate non-persistable
-    ancestors (which can't be serialized and transmitted to the other process) and any
-    of their follow-up tasks (which need to be computed immediately after their
-    parents). It also maintains a copy of this subgraph with unnecessary data pruned;
-    these "stripped" TaskStates can be safely serialized with cloudpickle and sent to
-    the other process.
+    should be run along with it. Task values can only be shared across
+    processes if they're persistable, so if we want to run a task remotely, we also
+    need to run all of its non-persistable dependencies (and their non-persistable
+    dependencies, and so on). Thus, this subgraph is constructed by starting from the
+    target TaskState and traversing outward through dependency and followup tasks,
+    stopping at any other persistable followup TaskStates. The final subgraph will
+    contain:
+
+    - A collection of `target_states` -- the original TaskState and any other
+      persistable followup states which need to be computed as well.
+    - A set of `external_dependency_states` --  any persistable TaskStates that need to
+      already be completed and persisted before this graph can be run.
+    - Any non-persistable TaskStates in between, which will also need to be computed.
+
+    This class also provides other fields and methods for determining if and how the
+    subgraph can be run. In particular, the `strip_states` method will make return a
+    copy of the task graph with unnecessary states and attributes removed, allowing
+    it to be serialized and transmitted to another process.
     """
 
-    def __init__(self, target_state, bootstrap):
-        self._bootstrap = bootstrap
+    def __init__(self, target_state):
+        self.all_states = set()
 
-        self._stripped_states_by_task_key = {}
-        self.persistable_but_not_persisted_states = set()
-        self.non_serializable_stripped_states = set()
-        self.stripped_states_with_aip_task_configs = set()
+        self.external_dependency_states = set()
+        self.target_states = set()
 
-        self._strip_state(target_state)
+        self.states_with_aip_task_configs = set()
+        self.non_serializable_states = set()
 
-    def _strip_state(self, original_state):
-        task_key = original_state.task_key
-        if task_key in self._stripped_states_by_task_key:
-            return self._stripped_states_by_task_key[task_key]
+        def add_state_and_traverse(state, is_dependency=False):
+            if state in self.all_states:
+                return
+            self.all_states.add(state)
 
-        # Make a copy of the TaskState, which we'll strip down to make it easier to
-        # serialize.
-        # (This is a shallow copy, so we'll make sure to avoid mutating any of its
-        # member variables.)
-        stripped_state = copy.copy(original_state)
-        self._stripped_states_by_task_key[task_key] = stripped_state
+            if state.should_persist:
+                assert len(state.followup_states) == 0
+                if is_dependency:
+                    self.external_dependency_states.add(state)
+                    return
+                else:
+                    self.target_states.add(state)
 
-        # Strip out data cached in memory -- we can't necessarily pickle it, so
-        # we need to get rid of it before trying to transmit this state to another
-        # process.
-        stripped_state._result = None
+            if state.func_attrs.aip_task_config is not None:
+                self.states_with_aip_task_configs.add(state)
+            if not state.task.can_be_serialized:
+                self.non_serializable_states.add(state)
 
-        # These fields are picklable, but only needed for cache setup and
-        # initialization.
-        if stripped_state.is_initialized:
-            stripped_state.case_key = None
+            for dep_state in state.dep_states:
+                add_state_and_traverse(dep_state, is_dependency=True)
+            for followup_state in state.followup_states:
+                add_state_and_traverse(followup_state)
 
-        if stripped_state.should_persist:
-            assert len(stripped_state.followup_states) == 0
+        add_state_and_traverse(target_state)
 
-            if stripped_state.is_cached_persistently:
+    def strip_states(self, states):
+        """
+        Returns copies of the provided TaskStates with any unnecessary state and
+        ancestors "stripped" off; these copies can be safely transmitted to another
+        process for computation.
+        """
+
+        stripped_states_by_task_key = {}
+
+        def strip_state(original_state):
+            """Returns a stripped copy of a TaskState."""
+
+            task_key = original_state.task_key
+            if task_key in stripped_states_by_task_key:
+                return stripped_states_by_task_key[task_key]
+
+            assert original_state in self.all_states
+            assert original_state not in self.non_serializable_states
+
+            # Make a copy of the TaskState, which we'll strip down to make it
+            # easier to serialize.
+            # (This is a shallow copy, so we'll make sure to avoid mutating any of
+            # its member variables.)
+            stripped_state = copy.copy(original_state)
+            stripped_states_by_task_key[task_key] = stripped_state
+
+            # Strip out data cached in memory -- we can't necessarily pickle it, so
+            # we need to get rid of it before trying to transmit this state to
+            # another process.
+            stripped_state._result = None
+
+            # External dependency states are expected to be already completed, so we
+            # don't need to include their task information or any of their dependencies.
+            if original_state in self.external_dependency_states:
                 stripped_state.task = None
                 stripped_state.func_attrs = None
                 stripped_state.dep_states = []
 
+            # Otherwise, we'll recursively strip all the dependency states as well.
             else:
-                self.persistable_but_not_persisted_states.add(original_state)
+                stripped_state.dep_states = [
+                    strip_state(dep_state) for dep_state in original_state.dep_states
+                ]
 
-        if stripped_state.task is not None:
-            if not stripped_state.task.can_be_serialized:
-                self.non_serializable_stripped_states.add(stripped_state)
-            if stripped_state.func_attrs.aip_task_config is not None:
-                self.stripped_states_with_aip_task_configs.add(stripped_state)
+            # We also strip and include any followup states.
+            stripped_state.followup_states = [
+                strip_state(followup_state)
+                for followup_state in original_state.followup_states
+            ]
 
-        stripped_state.dep_states = [
-            self._strip_state(dep_state) for dep_state in stripped_state.dep_states
-        ]
-        stripped_state.followup_states = [
-            self._strip_state(followup_state)
-            for followup_state in stripped_state.followup_states
-        ]
+            return stripped_state
 
-        return stripped_state
-
-    def get_stripped_state(self, original_state):
-        assert original_state in self.persistable_but_not_persisted_states
-        return self._stripped_states_by_task_key[original_state.task_key]
+        return [strip_state(state) for state in states]
 
     @property
     def all_states_can_be_serialized(self):
-        return len(self.non_serializable_stripped_states) == 0
+        return len(self.non_serializable_states) == 0
 
     @property
     def distinct_aip_task_configs(self):
         return set(
             state.func_attrs.aip_task_config
-            for state in self.stripped_states_with_aip_task_configs
+            for state in self.states_with_aip_task_configs
         )
