@@ -280,6 +280,7 @@ def test_code_hasher():
         "1",
         "123",
         "1.23",
+        "invalid utf8 \udc80",
         0,
         1,
         123,
@@ -557,6 +558,15 @@ def test_class_references():
     check_hash_equivalence([[type(1), type(2)], [type("str1"), type("str2")]])
 
 
+def test_same_class_different_module():
+    from . import helpers
+
+    class EmptyClass:
+        pass
+
+    check_hash_equivalence([[EmptyClass, helpers.EmptyClass]])
+
+
 def test_changes_in_references():
     v = 10
 
