@@ -4,7 +4,6 @@ import json
 
 import bionic as bn
 from bionic import interpret
-from bionic.gcs import get_gcs_fs_without_warnings
 from bionic.utils.urls import (
     path_from_url,
     is_file_url,
@@ -20,8 +19,9 @@ class CacheTester:
     in terms of changes between states.
     """
 
-    def __init__(self, flow, tier=["local", "cloud"]):
+    def __init__(self, flow, tier=["local", "cloud"], gcs_fs=None):
         self.flow = flow
+        self.gcs_fs = gcs_fs
         self._old_entries = set()
 
         self._tiers = interpret.str_or_seq_as_list(tier)
@@ -61,7 +61,7 @@ class CacheTester:
         ]
 
     def _validate_entry(self, entry):
-        artifact_bytes = read_bytes_from_url(entry.artifact_url)
+        artifact_bytes = read_bytes_from_url(entry.artifact_url, self.gcs_fs)
         value = json.loads(artifact_bytes)
         assert value == self.flow.get(entry.entity)
 
@@ -77,24 +77,25 @@ class CacheTester:
         # instead.)
         # TODO Hmm, is the above true? On closer inspection, it looks like artifact URLs
         # are derelativized right away when we load the metadata YAML.
-        metadata_str = read_bytes_from_url(entry.metadata_url).decode("utf-8")
+        metadata_str = read_bytes_from_url(entry.metadata_url, self.gcs_fs).decode(
+            "utf-8"
+        )
         assert entry.entity in metadata_str
 
 
-def read_bytes_from_url(url):
+def read_bytes_from_url(url, gcs_fs):
     """Reads the contents of a URL and returns them as a bytes object."""
 
     if is_file_url(url):
         path = path_from_url(url)
         return path.read_bytes()
     elif is_gcs_url(url):
-        gcs_fs = get_gcs_fs_without_warnings()
         return gcs_fs.cat_file(url)
     else:
         raise AssertionError(f"Unexpected scheme in URL: {url}")
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture
 def preset_flow(builder):
     builder.assign("x", 2)
     builder.assign("y", 3)
@@ -191,6 +192,52 @@ def test_flow_handles_delete_gracefully(builder):
     flow.get("c")
 
 
+@pytest.mark.needs_parallel
+@pytest.mark.allows_parallel
+def test_flow_handles_intermediate_deletion_in_parallel(builder, make_counter):
+    builder.assign("x", 2)
+
+    @builder
+    @bn.persist(False)
+    def y(x):
+        return x + 1
+
+    @builder
+    def z(y):
+        return y + 1
+
+    flow = builder.build()
+
+    assert flow.get("x") == 2
+    assert flow.get("y") == 3
+
+    for entry in flow.cache.get_entries():
+        entry.delete()
+
+    assert flow.get("z") == 4
+
+
+@pytest.mark.allows_parallel
+def test_flow_handles_partial_tuple_deletion(builder, make_counter):
+    builder.assign("x", 2)
+
+    @builder
+    @bn.outputs("y", "z")
+    def _(x):
+        return x - 1, x + 1
+
+    flow = builder.build()
+
+    assert flow.get("y") == 1
+    assert flow.get("z") == 3
+
+    (z_entry,) = [entry for entry in flow.cache.get_entries() if entry.entity == "z"]
+    z_entry.delete()
+
+    assert flow.get("y") == 1
+    assert flow.get("z") == 3
+
+
 def test_delete_artifact_with_multiple_metadata_files(builder):
     builder.assign("a", 1)
 
@@ -255,7 +302,7 @@ def test_delete_artifact_with_multiple_metadata_files(builder):
 # the GCS setup/teardown into `autouse` fixtures that are directly activated by the GCS
 # command line flag.
 @pytest.mark.needs_gcs
-def test_cache_on_gcs(gcs_builder):
+def test_cache_on_gcs(gcs_builder, gcs_fs):
     builder = gcs_builder
 
     builder.assign("a", 1)
@@ -270,9 +317,9 @@ def test_cache_on_gcs(gcs_builder):
 
     flow = builder.build()
 
-    local_tester = CacheTester(flow, tier="local")
-    cloud_tester = CacheTester(flow, tier="cloud")
-    total_tester = CacheTester(flow, tier=["local", "cloud"])
+    local_tester = CacheTester(flow, tier="local", gcs_fs=gcs_fs)
+    cloud_tester = CacheTester(flow, tier="cloud", gcs_fs=gcs_fs)
+    total_tester = CacheTester(flow, tier=["local", "cloud"], gcs_fs=gcs_fs)
 
     local_tester.expect_zero_entries()
     cloud_tester.expect_zero_entries()
