@@ -1,36 +1,7 @@
-import re
-
-import pytest
-
 import logging
 import threading
-import bionic as bn
 
-
-class LogChecker:
-    def __init__(self, caplog):
-        self._caplog = caplog
-
-    def expect_all(self, *expected_messages):
-        actual_messages = self._pop_messages()
-        assert set(actual_messages) == set(expected_messages)
-        self._caplog.clear()
-
-    def expect_regex(self, *expected_patterns):
-        actual_messages = self._pop_messages()
-        for pattern in expected_patterns:
-            assert any(re.fullmatch(pattern, message) for message in actual_messages)
-
-    def _pop_messages(self):
-        messages = [record.getMessage() for record in self._caplog.records]
-        self._caplog.clear()
-        return messages
-
-
-@pytest.fixture(scope="function")
-def log_checker(caplog):
-    caplog.set_level(logging.INFO)
-    return LogChecker(caplog)
+import pytest
 
 
 @pytest.mark.allows_parallel
@@ -131,46 +102,3 @@ def test_log_unpickleable_value(builder, log_checker):
         "Logging unpickleable class: Cannot pickle me",
         "Computed   log_unpickleable_value()",
     )
-
-
-@pytest.mark.no_parallel
-@pytest.mark.needs_aip
-def test_log_aip(aip_builder, log_checker):
-    builder = aip_builder
-
-    builder.assign("x", 1)
-
-    @builder
-    @bn.aip_task_config("n1-standard-4")
-    def x_plus_one(x):
-        return x + 1
-
-    @builder
-    @bn.aip_task_config("n1-standard-8")
-    def x_plus_two(x_plus_one):
-        return x_plus_one + 1
-
-    flow = builder.build()
-
-    assert flow.get("x_plus_one") == 2
-
-    log_checker.expect_regex(
-        r"Accessed   x\(x=1\) from definition",
-        r"Uploading x\(x=1\) to GCS ...",
-        r"Staging task .* at gs://.*",
-        r"Submitting .*x_plus_one.*",
-        r"Started task on AI Platform: https://console.cloud.google.com/ai-platform/jobs/.*",
-        r"Computed   x_plus_one\(x=1\) using AIP",
-    )
-
-    assert flow.get("x_plus_two") == 3
-
-    # Unfortunately, the string "Loaded     x_plus_one(x=1) from disk cache" is
-    # printed in AIP logs but not locally.
-
-    log_checker.expect_regex(
-        r"Staging task .* at gs://.*",
-        r"Submitting .*x_plus_two.*",
-        r"Started task on AI Platform: https://console.cloud.google.com/ai-platform/jobs/.*",
-        r"Computed   x_plus_two\(x=1\) using AIP",
-    )
