@@ -1,5 +1,6 @@
 import pytest
 import shutil
+import sys
 
 from .generate_test_compatibility_cache import Harness, CACHE_TEST_DIR
 
@@ -25,12 +26,22 @@ def older_serialized_cache_harness(make_counter, tmp_path):
 # To renegerate cache, run the following command from bionic/ dir
 #   `python -m tests.test_flow.generate_test_compatibility_cache`
 def test_caching_compatibility(older_serialized_cache_harness):
-    flow = older_serialized_cache_harness.flow
-    assert flow.get("xy") == 6
-    assert flow.get("yz") == 12
-    assert flow.get("xy_plus_yz") == 18
+    # The auto-versioned flow generates different bytecode hashes on Python 3.6
+    # compared to 3.7 and 3.8 This is because Python 3.7 added new
+    # bytecode instructions like LOAD_METHOD and uses them widely.
+    # We skip the auto-versioned flow for Python 3.6 and only test it on
+    # Python 3.7+.
+    if sys.version_info < (3, 7):
+        flows = [older_serialized_cache_harness.manual_flow]
+    else:
+        flows = older_serialized_cache_harness.flows
 
-    # assert that no methods were called
-    assert older_serialized_cache_harness.xy_counter.times_called() == 0
-    assert older_serialized_cache_harness.yz_counter.times_called() == 0
-    assert older_serialized_cache_harness.xy_plus_yz_counter.times_called() == 0
+    for flow in flows:
+        assert (
+            flow.get("total_sum") == older_serialized_cache_harness.EXPECTED_TOTAL_SUM
+        )
+
+        # Assert that no methods were called.
+        assert older_serialized_cache_harness.lowercase_sum_counter.times_called() == 0
+        assert older_serialized_cache_harness.uppercase_sum_counter.times_called() == 0
+        assert older_serialized_cache_harness.total_sum_counter.times_called() == 0
