@@ -5,9 +5,9 @@ from textwrap import dedent
 from random import Random
 
 from bionic.exception import CodeVersioningError
-from bionic.persistence import FakeCloudStore
 from bionic import interpret
 import bionic as bn
+from ..helpers import import_code
 
 
 class SimpleFlowModel:
@@ -165,19 +165,19 @@ class SimpleFlowModel:
         }
 
         e = entity
-        exec(
+        import_code(
             dedent(
                 f"""
             @builder
             @bn.changes_per_run({e.is_nondeterministic})
-            @bn.version(major={e.major_version}, minor={e.minor_version})
+            @bn.version_no_warnings(major={e.major_version}, minor={e.minor_version})
             def {name}({', '.join(e.dep_names)}):
                 noop_func({e.nonfunc_value})
                 record_call("{name}")
                 return {' + '.join([str(e.func_value)] + e.dep_names)}
             """
             ),
-            vars_dict,
+            vars_dict=vars_dict,
         )
 
         self._update_flow()
@@ -386,10 +386,8 @@ class Fuzzer:
 
 
 @pytest.fixture(scope="function")
-def fuzzer(builder, make_list, tmp_path):
-    fake_cloud_store = FakeCloudStore(str(tmp_path / "BNTESTDATA_FAKE_CLOUD"))
-    builder.set("core__persistent_cache__cloud_store", fake_cloud_store)
-    return Fuzzer(builder, make_list)
+def fuzzer(fake_gcs_builder, make_list, tmp_path):
+    return Fuzzer(fake_gcs_builder, make_list)
 
 
 foreach_mode = pytest.mark.parametrize("versioning_mode", ["manual", "assist", "auto"])
